from pathlib import Path

OTTER_PACKAGE = "odoo-otter"
OTTER_VERSION = "0.0.2"
OTTER_DESCRIPTION = "Otter - The Odoo time tracker"
OTTER_DIR = str(Path.home()) + "/.otter"
OTTER_CFG = f"{OTTER_DIR}/config.json"
OTTER_DB = f"{OTTER_DIR}/otter.db"

DOTS = '…'
SEPARATOR = ' | '

FORMAT_DATE = "%Y-%m-%d"
FORMAT_TIME = "%H:%M"
FORMAT_DATETIME = f"{FORMAT_DATE} {FORMAT_TIME}"



import logging

import requests

from otter.api import *
from otter.config import *
from otter.const import OTTER_CFG
from otter.database.query import *
from otter.errors import *
from otter.odoo import rest
from otter.util.cli import *
from otter.util.date import *


# logging.basicConfig(
#     level=logging.INFO,
#     # format="%(asctime)s [%(levelname)s] %(message)s",
#     handlers=[
#         logging.StreamHandler(sys.stdout)
#     ]
# )
# import validators
# or https://stackoverflow.com/a/7160778/4278102


def choose_project_task_id(args):
    limit = 9  # maybe from config

    records = query_records_most_recent(limit)

    if len(records) == 0:
        raise NoProjectTasksFoundError()

    # print("0: Search")
    for i, record in enumerate(records):
        print(f"{i + 1}: {record.project_task_abbrev(3)}")

    choice = int(input("?: "))

    if choice > limit:
        raise InvalidChoiceError()

    project_task_id = None

    if choice > 0:
        project_task_id = records[choice - 1].project_task_id

    # else:
    #     input("Project: ")
    #     input("Task: ")

    assert project_task_id is not None
    return project_task_id


def start(args):
    session = database.session()
    time_start = time_string_to_datetime(args.time)
    project_task_id = choose_project_task_id(args)
    record = get_or_create_record(project_task_id)

    try:
        # eiter stop it or use it
        record_time_slice = query_record_time_slice_active(session=session)

        # TODO: If date is not work date, ask user to stop it first

        # does running task matches selected one?
        if record_time_slice.record.project_task_id == project_task_id:
            print("Work on project task '{}' is already running since {}"
                  .format(record_time_slice.record.project_task_abbrev(),
                          record_time_slice.time_start_str))
            sys.exit(0)

        # if not, stop old one... TODO: Reuse stop here
        record_time_slice.time_stop = time_start
        session.commit()

        print("Stopped work on project task '{}' at {}"
              .format(record_time_slice.record.project_task_abbrev(),
                      record_time_slice.time_stop_str))

    except NoResultFound:
        pass

    # logging.info(f"Create new otter_record_time_slice's")
    record_time_slice = add_record_time_slice(record.id, time_start=time_start, session=session)
    session.commit()
    session.flush(record_time_slice)

    print("Started work on project task '{}' at {}"
          .format(record_time_slice.record.project_task_abbrev(),
                  record_time_slice.time_start_str))


def stop(args):
    session = database.session()
    time_stop = time_string_to_datetime(args.time)

    try:
        record_time_slice = query_record_time_slice_active(session=session)
        record_time_slice.time_stop = time_stop
        session.commit()

        print("Stopped work on project task '{}' at {}"
              .format(record_time_slice.record.project_task_abbrev(),
                      record_time_slice.time_stop_str))

    except NoResultFound:
        raise NoRecordActiveError()


def describe(args):
    session = database.session()
    description = args.description

    try:
        record_time_slice = query_record_time_slice_active(session=session)
        add_record_part(description, record_time_slice.record_id, session)
        session.commit()

        print("Tracked work '{}' for task '{}'"
              .format(description, record_time_slice.record.project_task_abbrev()))

    except NoResultFound:
        raise NoRecordActiveError()


def show(args):
    # Also offer finding by weekdays, see https://stackoverflow.com/q/319426/4278102
    session = database.session()
    query_date = work_date(date_string_to_date(args.date))
    records = get_records(query_date, session=session)

    if len(records) == 0:
        raise NoRecordForDateError(args.date)

    overall_duration = 0.0

    print("{}, {}:"
          .format(date_to_weekday_string(query_date),
                  date_to_date_string(query_date)))

    for record in records:
        overall_duration += record.duration
        print()
        print(record)

    h, m = divmod(round(overall_duration * 60), 60)

    print()
    print(f"[{h:02d}:{m:02d}] Overall")


def status(args):
    try:
        record = query_record_active()

        print("Active task:")
        print()
        print(record)

    except NoResultFound:
        raise NoRecordActiveError()


def sync(args):
    # Maybe transform into flag '--include-today'
    query_date = date_string_to_date(args.date)

    print("Syncing remote Odoo projects...", end=' ')
    sync_remote_projects()
    print("Done.")

    print("Syncing remote Odoo project tasks...", end=' ')
    sync_remote_project_tasks()
    print("Done.")

    print("Syncing local Otter records...", end=' ')
    sync_local_records(query_date)
    print("Done.")

    print("Syncing remote Odoo records...", end=' ')
    sync_remote_records()
    print("Done.")


def login(args):
    cfg = deserialize()

    url = input_default("URL", cfg['url'] if 'url' in cfg else None)
    url = url.rstrip('/')
    cfg['url'] = url
    serialize(cfg)

    # if not validators.url(url):
    #     sys.exit(1)

    databases = rest.get_databases_json(url)
    odoo_db = input_select_list_default("Database", databases, cfg['db'] if 'db' in cfg else None)
    cfg['db'] = odoo_db
    serialize(cfg)

    username = input_default("Username", cfg['user'] if 'user' in cfg else None)
    cfg['user'] = username
    serialize(cfg)

    password = args.password
    session = cfg['session'] if 'session' in cfg else None

    # TODO: Check if session still active
    # if session is not None:
    #     print("Already logged in as user '" + username + "'")
    #     sys.exit(0)

    if password is None:
        password = getpass.getpass(prompt='Password: ', stream=None)

    (uid, session) = rest.login(url, odoo_db, username, password)

    # # Contains all info like cookie expiry date
    # # Convert key-value String to dictionary using dict() + generator expression + split() + map()
    # auth_cookies = dict(map(str.strip, auth_cookie.split('=', 1)) \
    #     for auth_cookie in login_response.headers['set-cookie'].split('; ') if '=' in auth_cookie)

    cfg['uid'] = uid
    cfg['session'] = session
    serialize(cfg)
    print("Successfully logged in")


def logout(args):
    cfg = deserialize()
    user = cfg['user'] if 'user' in cfg else None
    if 'session' in cfg:
        del cfg['session']
        serialize(cfg)
        print("Successfully logged out as user '" + user + "'")
    else:
        print("Already logged out")



# based on https://realpython.com/python-sqlite-sqlalchemy/#using-flat-files-for-data-storage
from datetime import datetime

import sqlalchemy
from sqlalchemy import Column, String, Boolean, Integer, Float, Date, DateTime, ForeignKey, select, func, cast
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref, object_session

from otter.database import engine
# from otter.database import session as sess
from otter.util.cli import project_task_abbrev
from otter.util.date import datetime_to_time_string, date_string_to_date

Base = declarative_base()


# cached Odoo objects


# class BaseMixin(Base):
#     @classmethod
#     def create(cls, **kw):
#         obj = cls(**kw)
#         # TODO: That's so ugly
#         session = sess()
#         session.add(obj)
#         session.commit()


class Project(Base):
    __tablename__ = "project"

    odoo_id = Column(Integer, primary_key=True, autoincrement=False)
    name = Column(String)
    active = Column(Boolean)
    project_tasks = relationship("ProjectTask", backref=backref("project"))

    @classmethod
    def from_odoo_json(cls, json):
        return cls(
            odoo_id=json['id'],
            name=json['name'],
            active=json['active']
        )


class ProjectTask(Base):
    __tablename__ = "project_task"

    odoo_id = Column(Integer, primary_key=True, autoincrement=False)
    name = Column(String)
    active = Column(Boolean)
    project_id = Column(Integer, ForeignKey("project.odoo_id"))
    records = relationship("Record", backref=backref("project_task"))

    @classmethod
    def from_odoo_json(cls, json):
        return cls(
            odoo_id=json['id'],
            name=json['name'],
            active=json['active'],
            project_id=json['project_id'][0]
        )


class Record(Base):
    __tablename__ = "record"

    id = Column(Integer, primary_key=True, autoincrement=True)
    odoo_id = Column(Integer, unique=True)
    date = Column(Date, nullable=False)
    name = Column(String)
    unit_amount = Column(Float)
    project_task_id = Column(Integer, ForeignKey("project_task.odoo_id"))
    record_track_items = relationship("RecordTrackItem", backref=backref("record"),
                                      cascade="all, delete, delete-orphan")
    record_time_slices = relationship("RecordTimeSlice", backref=backref("record"),
                                      cascade="all, delete, delete-orphan")

    def __repr__(self):
        hh_mm = f"[{self.duration_str}]"
        return (
            f"{hh_mm} {self.project_task_abbrev(len(hh_mm)+1)}\n" +
            f"        {self.description}"
        )

    @classmethod
    def get_id_by_odoo_id(cls, odoo_id, session):
        try:
            record_id, = session.query(Record.id).filter_by(odoo_id=odoo_id).one()
            return record_id
        except NoResultFound:
            return None

    @classmethod
    def from_odoo_json(cls, local_id, json):
        return cls(
            id=local_id,
            odoo_id=json['id'],
            date=date_string_to_date(json['date']),
            name=json['name'],
            unit_amount=json['unit_amount'],
            project_task_id=json['task_id'][0]
        )

    @property
    def duration(self):
        if self.unit_amount is not None:
            return self.unit_amount
        return (
            object_session(self)
            .scalar(
                select(func.sum((
                    cast(func.coalesce(func.julianday(RecordTimeSlice.time_stop), func.julianday(datetime.now()))
                         - func.julianday(RecordTimeSlice.time_start), sqlalchemy.Float) * 24
                )))
                .where(RecordTimeSlice.record_id == self.id)
            )
        )

    @property
    def duration_str(self):
        if self.duration is not None:
            h, m = divmod(round(self.duration * 60), 60)
            return f"{h:02d}:{m:02d}"
        return None

    @property
    def description(self):
        if self.name is not None:
            return self.name
        return (
            object_session(self)
            .scalar(
                select(func.group_concat(RecordTrackItem.name, ', '))
                .where(RecordTrackItem.record_id == self.id)
            )
        )

    @property
    def project_id(self):
        return self.project_task.project_id

    @property
    def project_name(self):
        return self.project_task.project.name

    @property
    def project_task_name(self):
        return self.project_task.name

    def project_task_abbrev(self, offset_length=0):
        return project_task_abbrev(self.project_name, self.project_task_name, offset_length)

    def odoo_create_json(self):
        return {
            'date': self.date,
            'project_id': self.project_id,
            'task_id': self.project_task_id,
            'duration': self.unit_amount,
            'names': self.name
        }


# Otter objects


class RecordTrackItem(Base):
    __tablename__ = "record_track_item"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    record_id = Column(Integer, ForeignKey("record.id"), nullable=False)


class RecordTimeSlice(Base):
    __tablename__ = "record_time_slice"

    id = Column(Integer, primary_key=True, autoincrement=True)
    # the start time may never be null while the stop time is null until it the work got stopped
    time_start = Column(DateTime, nullable=False)
    time_stop = Column(DateTime, nullable=True)
    record_id = Column(Integer, ForeignKey("record.id"), nullable=False)

    @property
    def time_start_str(self):
        return datetime_to_time_string(self.time_start)

    @property
    def time_stop_str(self):
        return datetime_to_time_string(self.time_stop)


Base.metadata.create_all(engine())



import json

from otter.const import OTTER_CFG
from otter.util.serialize import from_json


def serialize(items):
    with open(OTTER_CFG, 'w') as file:
        json.dump(items, file, indent=2)


def deserialize():
    with open(OTTER_CFG, 'r') as f:
        string = f.read()

        if len(string) == 0:
            return dict()

        return from_json(string)



import requests

from otter.config import deserialize


def cookies():
    config = deserialize()
    session = config['session']
    return requests.utils.cookiejar_from_dict({"session_id":session})
